from Globals import HTMLFile, MessageDialog, Persistent
import App
import Acquisition
from OFS.SimpleItem import Item
from OFS.Folder import Folder

import string
import time
import sys
import types
import StringIO
import mimetypes
import re
import rfc822
import os
import urllib
import mimify

try:
    import DateTime
except: pass


from ImapUtil import CreateStruct, makemap, Internaldate2tuple
import Imap
import MimeTools

                
class MsgPart:
	def __init__(self, BodyPart, BodyType, BodySubtype, BodyParameterList=None, BodyId=None, BodyDescription=None, BodyEncoding=None, BodySize=None, BodyEnvelope=None, BodyLines=None, BodyMD5=None, BodyDisposition=None, BodyLanguage=None, BodyHeader=None, BodyText=None, BodyMime=None, BodyAttachment=None, BodyContentType=None):
        	self.BodyPart=BodyPart
		self.BodyType=BodyType
        	self.BodySubtype=BodySubtype
                self.BodyParameterList=BodyParameterList
		self.BodyId=BodyId
	        self.BodyDescription=BodyDescription
        	self.BodyEncoding=BodyEncoding
                self.BodySize=BodySize
	        self.BodyEnvelope=BodyEnvelope
                self.BodyLines=BodyLines
	        self.BodyMD5=BodyMD5
        	self.BodyDisposition=BodyDisposition
                self.BodyLanguage=BodyLanguage
                self.BodyHeader=BodyHeader
                self.BodyText=BodyText
                self.BodyMime=BodyMime
                self.BodyAttachment=BodyAttachment
                self.BodyContentType=BodyContentType

class MsgEnvelope:
	def __init__(self, EnvDate, EnvSubject, EnvFrom, EnvSender, EnvReplyTo, EnvTo, EnvCc, EnvBcc, EnvInReplyTo, EnvMessageId):
		self.EnvDate=EnvDate
                self.EnvSubject=EnvSubject
                self.EnvFrom=EnvFrom
                self.EnvSender=EnvSender
                self.EnvReplyTo=EnvReplyTo
                self.EnvTo=EnvTo
                self.EnvCc=EnvCc
                self.EnvBcc=EnvBcc
                self.EnvInReplyTo=EnvInReplyTo
                self.EnvMessageId=EnvMessageId

class MsgAddress:
	def __init__(self, AddrPersonalName, AddrAtDomainList, AddrMailboxName, AddrHostName):
		self.AddrPersonalName=AddrPersonalName
                self.AddrAtDomainList=AddrAtDomainList
                self.AddrMailboxName=AddrMailboxName
                self.AddrHostName=AddrHostName


class MsgParameter:
	def __init__(self, ParameterName, ParameterValue):
		self.ParameterName=ParameterName
		self.ParameterValue=ParameterValue

class PartDisposition:
	def __init__(self, PartDispositionType, PartDispositionAttr, PartDispositionAttrMap=None):
		self.PartDispositionType=PartDispositionType
                self.PartDispositionAttr=PartDispositionAttr
                if PartDispositionAttrMap:
	                self.PartDispositionAttrMap=PartDispositionAttrMap
                else:
                	if PartDispositionAttr:
	                	self.PartDispositionAttrMap={}
        	        	for i in PartDispositionAttr:
                	        	self.PartDispositionAttrMap[i.ParameterName]=i
                        	

class MsgMessage:
	def __init__(self, MsgUid, MsgFlags, MsgInternalDate, MsgRfc822Size, MsgEnvelope, MsgParts, MsgText, MsgTextPart=None):
		self.MsgUid=MsgUid
                self.MsgFlags=MsgFlags
                self.MsgInternalDate=MsgInternalDate
                self.MsgRfc822Size=MsgRfc822Size
                self.MsgEnvelope=MsgEnvelope
                self.MsgParts=MsgParts
                self.MsgText=MsgText
                self.MsgTextPart=MsgTextPart


class ImapMessage(Item,Acquisition.Implicit):
	'''IMAP Message'''


	meta_type = 'ImapMessage'
	icon='misc_/ImapClient/Message'

	manage_options=(
		{'label':'Message','icon':icon,'action':'manage_main','target':'manage_main'},
		{'label':'Attachments','icon':icon,'action':'manage_attachments','target':'manage_main'},
		{'label':'Info','icon':icon,'action':'manage_info','target':'manage_main'}
		)

	__ac_permissions__=(
		('View management screens',('manage_tabs','manage_main')),
		('Change permissions',('manage_access',)),
		('Change Message',('manage_edit',),('Manager',)),
		)


	def __init__(self, ID, Flags, FromName, FromAddress, ToName, ToAddress, Date, Subject, Size):
		self.UID=ID
		self.Flags=Flags
		self.FromName=FromName
		self.FromAddress=FromAddress
		self.ToName=ToName
		self.ToAddress=ToAddress
		self.Size=Size
		try:
			self.Date=self.makeDateTimeFromRfc822(Date)
		except:
			self.Date = Date
		try:
			self.Subject=mimify.mime_decode_header(Subject)
		except:
			self.Subject=Subject

		self.title = self.Subject
		self._message = None

	def bobobase_modification_time(self):
		return self.Date

	def get_size(self):
		return int(self.Size)
		
	def updateMessageFlags(self,Flags):
		self.Flags = Flags

	def getMessage(self,REQUEST=None):
		''' docstring '''
		if not self._message:
			self._message = self.getIMAPMsg()
		return self._message
		
	manage_main = HTMLFile('manage/message',globals())
	manage_attachments = HTMLFile('manage/message_attachments',globals())
	manage_info = HTMLFile('manage/message_info',globals())
	
	def makeAddressList(self,ST):
        	if ST:
	        	al=[]
                	for ADDR in ST:
                        	al.append(MsgAddress(ADDR[0],ADDR[1],ADDR[2],ADDR[3]))
                        return al
                else:
                	return [] ##used to be None
                        	

	def makeEnvelope(self, ST):
        	#print "ENVELOPESOURCE", ST
        	if ST:
                	return(MsgEnvelope(
                        	self.makeDateTimeFromRfc822(ST[0]),
                                ST[1],
                                self.makeAddressList(ST[2]),
                                self.makeAddressList(ST[3]),
                                self.makeAddressList(ST[4]),
                                self.makeAddressList(ST[5]),
                                self.makeAddressList(ST[6]),
                                self.makeAddressList(ST[7]),
                        	ST[8],
                                ST[9],
                        	))
                else:
                	return None

	def makePartDisposition(self, ST):
        	if ST:
                	l=[]
                        m={}
                        array=ST[1]
			for i in range(0,len(array),2):
				item=array[i]
                                value=array[i+1]
				l.append(MsgParameter(item, value))
                                m[item]=value
                	return PartDisposition(ST[0], l, m)
                else:
                	return None
        
	def makePartList(self, M, UID, Part, PL, ST, Depth):
        	MsgText=None
                MsgTextPart=None
                MultiParts=[]
        	partnr=0
		if not len(ST): return None
                
		while partnr<len(ST) and type(ST[partnr])==type([]):
                	MultiParts.append(ST[partnr])
                        partnr=partnr+1

                if partnr>=len(ST): return None
                if partnr>0:
                	"""Multipart Part"""
                        BodyType="MULTIPART"
                        BodySubtype=ST[partnr+0]                        
                        BodyParameterList=ST[partnr+1]
                        BodyDisposition=self.makePartDisposition(ST[partnr+2])
                        BodyLanguage=ST[partnr+3]
                        if Part=="":
                        	HeaderPart="RFC822.HEADER"
                        else:
                        	HeaderPart="BODY["+Part+".HEADER]"
	                tf=M.uid('FETCH', '%d'%(UID,), '%s'%(HeaderPart))
                       	ts=CreateStruct(tf[1])
                        tm=makemap(ts[0][1])
			BodyHeader=tm[HeaderPart]
                        BodyAttachment=None
			PL.append(MsgPart(
                        	BodyPart=Part,
		                BodyType=BodyType,
        	                BodySubtype=BodySubtype,
                                BodyContentType=BodyType+"/"+BodySubtype,
                	        BodyParameterList=BodyParameterList,
                        	#BodyId=BodyId,
	                        #BodyDescription=BodyDescription,
        	                #BodyEncoding=BodyEncoding,
                	        #BodySize=BodySize,
	                        #BodyEnvelope=BodyEnvelope,
                                #BodyLines=BodyLines,
	                        #BodyMD5=BodyMD5,
        	                BodyDisposition=BodyDisposition,
                	        BodyLanguage=BodyLanguage,
                                BodyHeader=BodyHeader,
                                #BodyText=BodyText,
                                BodyAttachment=BodyAttachment,
                                ))
                else:
                        offset=0
	                BodyType=string.upper(ST[0])
                        BodySubtype=string.upper(ST[1])
                        BodyParameterList=ST[2]
                        BodyId=ST[3]
                        BodyDescription=ST[4]
                        BodyEncoding=ST[5]
                        BodySize=ST[6]
                        BodyEncoding=string.lower(BodyEncoding)
                        BodyEnvelope=None
                        BodyStructure=None
                        BodyLines=None
                        if BodyType=="MESSAGE" and BodySubtype=="RFC822":
                        	try:
		                        BodyEnvelope=self.makeEnvelope(ST[7])
		                        BodyStructure=ST[7+1]
                	                BodyLines=ST[7+2]
                                except:
                                	pass
                                offset=3
                        if BodyType=="TEXT":
                                BodyLines=ST[7]
                                offset=1
                        
                        BodyMD5=ST[7+offset]
                        BodyDisposition=self.makePartDisposition(ST[8+offset])
                        BodyLanguage=ST[9+offset]

                        BodyText=None
                        BodyHeader=None
                        BodyMime=None
                        
                        #print
        	        #print BodyType, BodySubtype
                        #print
                        
        	        if (BodyType=='TEXT' #and BodySubtype=='PLAIN'
                        	or BodyType=="MESSAGE" and BodySubtype!="RFC822"
                        	):
                        	BodyPart="BODY["+Part+"]"
                        	if Part=="":
                                	HeaderPart="RFC822.HEADER"
		        	        if (BodyType=='TEXT' and BodySubtype=='PLAIN'):
                                		TextPart="RFC822.TEXT"
                                        else:
                                		TextPart="RFC822.TEXT"
                                else:
		        	        if (BodyType=='TEXT' and BodySubtype=='PLAIN'):
                                		TextPart="BODY["+Part+".TEXT]"
                                		HeaderPart="BODY["+Part+".HEADER]"
                                        else:
                                		HeaderPart="BODY["+Part+".HEADER]"
                                		TextPart="BODY["+Part+".TEXT]"
                                if HeaderPart!="":
	                		tf=M.uid('FETCH', '%d'%(UID,), '(%s %s)'%(HeaderPart, TextPart))
                                else:
	                		tf=M.uid('FETCH', '%d'%(UID,), '%s'%(TextPart))
                       		ts=CreateStruct(tf[1])
                               	tm=makemap(ts[0][1])
                                #print tm
                                if HeaderPart!="":
					BodyHeader=tm[HeaderPart]
                                #if BodyType=="MESSAGE" and BodySubtype=="RFC822":
                                #	BodyText=None
                                #else:
				BodyText=tm[TextPart]
                                if BodyText==None:
	                		tf=M.uid('FETCH', '%d'%(UID,), '%s'%(BodyPart))
	                       		ts=CreateStruct(tf[1])
        	                       	tm=makemap(ts[0][1])
					BodyText=tm[BodyPart]
                                #print "BODYENCODING", BodyEncoding
                                if BodyEncoding in ('quoted-printable', 'base64'):
					Input=StringIO.StringIO(BodyText)
                                	DecodedText=StringIO.StringIO()
                                        #print "DECODEING", BodyEncoding
					MimeTools.decode(Input, DecodedText, BodyEncoding)
                			BodyText=DecodedText.getvalue()
                                        
                                if not MsgText and BodySubtype=='PLAIN':#and Part=="" 
                                	MsgText=BodyText
                                        MsgTextPart=Part
                                #print "HEADER:", BodyHeader
                                #print "TEXT:", BodyText
                                BodyAttachment=None
                        elif BodyType=="MESSAGE" and BodySubtype=="RFC822":
                                if BodyEncoding and string.lower(BodyEncoding) in ('7bit','8bit','quoted-printable'):
        	                        BodyAttachment=None
                                else:
	                        	BodyAttachment=1
                        else:
                        	BodyAttachment=1
                        try:
                        	MimePart="BODY["+Part+".MIME]"
		                tf=M.uid('FETCH', '%d'%(UID,), '%s'%(MimePart))
        	               	ts=CreateStruct(tf[1])
                	        tm=makemap(ts[0][1])
                        	#print tm
				BodyMime=tm[MimePart]
                        except:
                        	pass

			PL.append(MsgPart(
                        	BodyPart=Part,
		                BodyType=BodyType,
        	                BodySubtype=BodySubtype,
                                BodyContentType=BodyType+"/"+BodySubtype,
                	        BodyParameterList=BodyParameterList,
                        	BodyId=BodyId,
	                        BodyDescription=BodyDescription,
        	                BodyEncoding=BodyEncoding,
                	        BodySize=BodySize,
	                        BodyEnvelope=BodyEnvelope,
                                BodyLines=BodyLines,
	                        BodyMD5=BodyMD5,
        	                BodyDisposition=BodyDisposition,
                	        BodyLanguage=BodyLanguage,
                                BodyHeader=BodyHeader,
                                BodyText=BodyText,
                                BodyMime=BodyMime,
                                BodyAttachment=BodyAttachment,
                                ))
                if Depth>0:
                	Depth=Depth-1
			if BodyType=="MESSAGE" and BodySubtype=="RFC822" and not BodyAttachment:
		        	self.makePartList(M, UID, Part, PL, BodyStructure, Depth)
                	partnr=0
			while partnr<len(MultiParts):
        	        	if Part!="":
                	        	ThisPart=Part+"."+str(partnr+1)
                        	else:
                        		ThisPart=str(partnr+1)
                                if not MsgText:
		                	MsgTextPart, MsgText=self.makePartList(M, UID, ThisPart, PL, ST[partnr], Depth)
                                else:
		                	self.makePartList(M, UID, ThisPart, PL, ST[partnr], Depth)
        	                partnr=partnr+1
                return MsgTextPart, MsgText 
                                
        def makeDateTimeFromInternaldate(self, tis):
        	if len(tis)>20:
			ti=Internaldate2tuple('INTERNALDATE "'+ tis + '"')
                else:
			ti=Internaldate2tuple('INTERNALDATE "'+ tis + ' +0000"')
		try:
			dt=None
			dt=DateTime.DateTime()
			dt=DateTime.DateTime(time.mktime(ti))
		except:
                	pass
                return dt
                        	
        def makeDateTimeFromRfc822(self, tis):
        	if len(tis)>25:
                	ti=rfc822.mktime_tz(rfc822.parsedate_tz(tis))
                else:
                	ti=time.mktime(rfc822.parsedate(tis))
		try:
			dt=None
			dt=DateTime.DateTime()
			dt=DateTime.DateTime(ti)
		except:
                	pass
                return dt
                        	
	def makeMessage(self, M, Depth, ST):
        	#print ST
        	MsgParts=[]
        	MsgUid=ST['UID']
                MsgFlags=ST['FLAGS']
                MsgInternalDate=self.makeDateTimeFromInternaldate(ST['INTERNALDATE'])
                MsgRfc822Size=['RFC822.SIZE']
                MsgEnvelope=self.makeEnvelope(ST['ENVELOPE'])
                try:
			MsgTextPart,MsgText=self.makePartList(M, MsgUid, "", MsgParts, ST['BODYSTRUCTURE'], Depth)
                except:
                	MsgTextPart,MsgText=0,"Sorry, Problem Retrieving Text"
                return( MsgMessage(
			MsgUid=MsgUid,
                	MsgFlags=MsgFlags,
                	MsgInternalDate=MsgInternalDate,
                	MsgRfc822Size=MsgRfc822Size,
                        MsgEnvelope=MsgEnvelope,
                	MsgParts=MsgParts,
                        MsgText=MsgText,
                        MsgTextPart=MsgTextPart,
                        ))

	def getPart(self, REQUEST, RESPONSE):
		''' official ?? '''
        	if REQUEST.has_key('Part'): Part=REQUEST['Part']
                else: Part="1"
        	if REQUEST.has_key('MimeEncoding'): MimeEncoding=string.lower(REQUEST['MimeEncoding'])
                else: MimeEncoding="base64"
        	if REQUEST.has_key('AttType'): AttType=string.lower(REQUEST['AttType'])
                else: AttType=""
        	if REQUEST.has_key('AttSubType'): AttSubType=string.lower(REQUEST['AttSubType'])
                else: AttSubType=""
        	if REQUEST.has_key('FileName'): FileName=string.lower(REQUEST['FileName'])
                else: FileName="ATT0001.DAT"
        	M=self.getImapConnection(self.getFolderFullName())
		E=M.uid('FETCH', '%s'%(self.UID), 'BODY[%s]'%(Part))
		ES=CreateStruct(E[1])
		EM=makemap(ES[0][1])
		input=StringIO.StringIO(EM['BODY[%s]'%(Part)])
                output=StringIO.StringIO()
		MimeTools.decode(input, output, MimeEncoding)
                decodeddata=output.getvalue()
		open("/tmp/att2","w").write(decodeddata)
                undecodeddata=input.getvalue()
		open("/tmp/att","w").write(undecodeddata)
                RESPONSE.setHeader('content-type', AttType+'/'+AttSubType)
                RESPONSE.setHeader('content-length', len(decodeddata))
                RESPONSE.setHeader('content-disposition', 'inline; filename="'+FileName+'"')
                return decodeddata
                
                
	def getPartRaw(self, Part, Output, REQUEST=None):
        	M=self.getImapConnection(self.getFolderFullName())
		E=M.uid('FETCH', '%s'%(self.UID), 'BODY[%s]'%(Part))
		ES=CreateStruct(E[1])
		EM=makemap(ES[0][1])
                Output.write(EM['BODY[%s]'%(Part)])
                return 1
                
                
	def getIMAPMsg(self, Depth=20, REQUEST=None, peek=0):
		st=[]
                M=self.getImapConnection(self.getFolderFullName(),rw=1)
                
		E=M.uid('FETCH', '%d'%(int(self.UID)), '(FLAGS INTERNALDATE RFC822.SIZE ENVELOPE BODYSTRUCTURE)')
		if E[1][0]!=None:
			ES=CreateStruct(E[1])
		else:
			return ()
                wasseen=0
		for mail in ES:
			sequence_number=mail[0]
                        mi=makemap(mail[1])
                        rv=self.makeMessage(M, Depth, mi)
	                if not peek:
                        	M.uid("STORE", self.UID, '+FLAGS.SILENT', '(\Seen)')
                        else:
                        	if not wasseen: M.uid("STORE", self.UID, '-FLAGS.SILENT', '(\Seen)')
                        return rv
	
	
